@@ -86,6 +86,11 @@ struct smtp_server_recipient {
 };
 ARRAY_DEFINE_TYPE(smtp_server_recipient, struct smtp_server_recipient *);
 
+/* Returns the original recipient path if available. Otherwise, it returns the
+   final path. */
+const struct smtp_address *
+smtp_server_recipient_get_original(struct smtp_server_recipient *rcpt);
+
 struct smtp_server_reply *
 smtp_server_recipient_get_reply(struct smtp_server_recipient *rcpt);
 bool smtp_server_recipient_is_replied(struct smtp_server_recipient *rcpt);
@@ -122,6 +127,15 @@ void smtp_server_recipient_remove_hook(
  * Transaction
  */
 
+enum smtp_server_trace_rcpt_to_address {
+	/* Don't add recipient address to trace header. */
+	SMTP_SERVER_TRACE_RCPT_TO_ADDRESS_NONE,
+	/* Add final recipient address to trace header. */
+	SMTP_SERVER_TRACE_RCPT_TO_ADDRESS_FINAL,
+	/* Add original recipient address to trace header. */
+	SMTP_SERVER_TRACE_RCPT_TO_ADDRESS_ORIGINAL,
+};
+
 enum smtp_server_transaction_flags {
 	SMTP_SERVER_TRANSACTION_FLAG_REPLY_PER_RCPT = BIT(0),
 };
@@ -160,8 +174,9 @@ void smtp_server_transaction_fail_data(
 	unsigned int status, const char *enh_code,
 	const char *fmt, va_list args) ATTR_FORMAT(5, 0);
 
-void smtp_server_transaction_write_trace_record(string_t *str,
-	struct smtp_server_transaction *trans);
+void smtp_server_transaction_write_trace_record(
+	string_t *str, struct smtp_server_transaction *trans,
+	enum smtp_server_trace_rcpt_to_address rcpt_to_address);
 
 /*
  * Callbacks
@@ -277,7 +292,8 @@ struct smtp_server_callbacks {
 
 	/* Protocol state events */
 	void (*conn_state_changed)(void *context,
-				   enum smtp_server_state newstate);
+				   enum smtp_server_state new_state,
+				   const char *new_args) ATTR_NULL(3);
 
 	/* Proxy data */
 	void (*conn_proxy_data_updated)(void *conn_ctx,
@@ -286,8 +302,12 @@ struct smtp_server_callbacks {
 	/* Connection */
 	int (*conn_start_tls)(void *conn_ctx,
 		struct istream **input, struct ostream **output);
+	/* Connection is disconnected. This is always called before
+	   conn_free(). */
 	void (*conn_disconnect)(void *context, const char *reason);
-	void (*conn_destroy)(void *context);
+	/* The last reference to connection is dropped, causing the connection
+	   to be freed. */
+	void (*conn_free)(void *context);
 
 	/* Security */
 	bool (*conn_is_trusted)(void *context);
@@ -364,9 +384,20 @@ struct smtp_server_settings {
 	bool auth_optional:1;
 	/* TLS security is required for this service */
 	bool tls_required:1;
+	/* The path provided to the MAIL command does not need to be valid. A
+	   completely invalid path will parse as <>. Paths that can still be
+	   fixed by splitting it on the last `@' yielding a usable localpart and
+	   domain, will be parsed as such. There are limits though; when the
+	   path is badly delimited or contains control characters, the MAIL
+	   command will still fail. The unparsed broken address will be
+	   available in the `raw' field of struct smtp_address for logging etc.
+	 */
+	bool mail_path_allow_broken:1;
 	/* The path provided to the RCPT command does not need to have the
 	   domain part. */
 	bool rcpt_domain_optional:1;
+	/* Don't include "(state=%s)" in the disconnection reason string. */
+	bool no_state_in_reason:1;
 };
 
 struct smtp_server_stats {
@@ -420,6 +451,10 @@ void smtp_server_connection_start(struct smtp_server_connection *conn);
    handling command input is held off until smtp_server_connection_resume() is
    called. */
 void smtp_server_connection_start_pending(struct smtp_server_connection *conn);
+/* Abort the connection prematurely (before it is started). */
+void smtp_server_connection_abort(struct smtp_server_connection **_conn,
+				  unsigned int status, const char *enh_code,
+				  const char *reason);
 
 /* Halt connection command input and idle timeout entirely. */
 void smtp_server_connection_halt(struct smtp_server_connection *conn);
@@ -447,7 +482,8 @@ int smtp_server_connection_data_chunk_add(struct smtp_server_cmd_ctx *cmd,
 	bool client_input);
 
 enum smtp_server_state
-smtp_server_connection_get_state(struct smtp_server_connection *conn);
+smtp_server_connection_get_state(struct smtp_server_connection *conn,
+				 const char **args_r) ATTR_NULL(2);
 const char *
 smtp_server_connection_get_security_string(struct smtp_server_connection *conn);
 struct smtp_server_transaction *
@@ -634,8 +670,19 @@ struct smtp_server_reply *
 smtp_server_reply_create_forward(struct smtp_server_command *cmd,
 	unsigned int index, const struct smtp_reply *from);
 
+void smtp_server_reply_set_status(struct smtp_server_reply *reply,
+				  unsigned int status, const char *enh_code)
+				  ATTR_NULL(3);
+unsigned int smtp_server_reply_get_status(struct smtp_server_reply *reply,
+					  const char **enh_code_r) ATTR_NULL(3);
+
 void smtp_server_reply_add_text(struct smtp_server_reply *reply,
 	const char *line);
+void smtp_server_reply_prepend_text(struct smtp_server_reply *reply,
+				    const char *text_prefix);
+void smtp_server_reply_replace_path(struct smtp_server_reply *reply,
+				    struct smtp_address *path, bool add);
+
 void smtp_server_reply_submit(struct smtp_server_reply *reply);
 void smtp_server_reply_submit_duplicate(struct smtp_server_cmd_ctx *_cmd,
 					unsigned int index,
